#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''h5_pytables.py: Old HDF5 funcionality'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import pandas as pd


class DataFile():
    """
    === HDF5 ===

    Using the HDF5 file format has its own possitive features. For one it reads
    and writes much faster then using the ASCII format. Also attributes
    (like the grid definition) can be saved with the file and pygeostat will
    autoload those attributes. All files for a single project can also be saved
    in the same file as separate tables and each table or subset of a table can
    be read in as needed.
    HDF5 file simple read example

    >>> datafl = gs.DataFile(flname='../data/oilsands_out.hdf5', dftype='hdf5',
    >>>                      hdf_key='gslib_data')

    To view the HDF5 header information (tables stored in the file)

    >>> datafl.store

    If you have a HDF5 file with multiple tables and you just want to read in the file information
    to view what tables are in the file and any attributes saved to the file you can do a header
    style only read

    >>> datafl = gs.DataFile(flname='../data/oilsands_out.hdf5', dftype='hdf5',
    >>>                      headeronly=True, hdf5_header=True)

    Then to see what tables are written in the hdf5 file

    >>> datafl.store
    """



    def retrieve_hdf5(self, hdf_key):
        '''Retrieve a new group from self.store and load it into the self.data location
        of the gs.DataFile()

        Note:
            This currently won't work if you don't have a self.store in the DataFile.
            You need to open a .h5 file with the header option in order to use this.

        Parameters:
            hdf_key (str): name of the group to load into self.data

        Examples:

        load the initial data file (say the 1st realization)

        >>> datafl = gs.DataFile(flname='sgsim.out', hdf5_header=True, hdf_key='Realization_1')
        >>> # Look at some summary statistics of the first realization
        >>> datafl.describe()

        Later you decide you want to look at the 4th realization so now you retrieve that data

        >>> datafl.retrieve_hdf5('Realization_4')
        >>> # Look at the new summary statistics for the 4th realization
        >>> datafl.describe()

        '''

        if self.store is None:
            raise Exception('There is no self.store - Please read in an hdf5 file with the header',
                            ' option')
        else:
            from ..data import readfile as readfile
            self.data = readfile(self.flname, fltype='hdf5', hdf_key=hdf_key)

        return self

        from .data import open_h5 as open_h5
        # read the file and import data and/or header
        if headeronly is True:
            if hdf5_header is True:
                self.store = open_h5(flname)
                self.data = pd.DataFrame()
            elif hdf5_header is False:
                self.data = readfile(self.flname, fltype=fltype, headeronly=headeronly,
                                     delimiter=delimiter, hdf_key=hdf_key, columns=columns)
        elif headeronly is False:
            if hdf5_header is True:
                self.data = readfile(self.flname, fltype=fltype, headeronly=headeronly,
                                     delimiter=delimiter, hdf_key=hdf_key, columns=columns)
                self.store = open_h5(flname)
            elif hdf5_header is False:
                self.data = readfile(self.flname, fltype=fltype, headeronly=headeronly,
                                     delimiter=delimiter, hdf_key=hdf_key, columns=columns)


        # If a hdf5 file was read in check to see if any file attributes were loaded
        if hasattr(self, 'store'):
            if hasattr(self.store.root._v_attrs, 'gridstr'):
                gridstr = self.store.root._v_attrs.gridstr
                from .data import GridDef as GridDef
                if self.griddef is not None:
                    print('Warning: Grid definition is being rewritten by griddef saved in h5 file')
                if gridstr == 'None':
                    print('Warning: unable to set grid definition due to empty grid string saved',
                          ' in h5 file')
                else:
                    self.griddef = GridDef(gridstr)
            # set attributes call

            self.attrs = self.store.root._v_attrs

def write_h5_pytables(data, flname, variables=None, hdf_key='gslib_data', mode='w', hdf_append=False, complevel=None, complib=None, griddef=None):
    """Writes out a HDF5 binary data file. This is useful for greatly increasing
    the read and write speads to memory which is a common bottleneck. There are
    some usage notes and warnings that should be reviewed.

    For more information on HDF5 - search youtube for some of the python hdf5 review videos.

    For a HDF5 file viewing software (outside of python) with a GUI interface check out
    the HDFView software at http://www.hdfgroup.org

    Notes:
        For out of the box best speed - use the default of no compression.

        If you want faster speed then reading and writing text (ie. gslib or csv) but
        also have some file size concerns then the best current out of the box option
        is using the "zlib" compression library.

        The "blosc" compression library is supposed to have both faster read/write speeds
        than no compression and it provides some compression. It produces faster read/write
        speeds by reorganizing the structure of the binary file in the same way the computer
        memory stores information. *Warning* Currently the "blosc" compression library used
        is the one packaged in pytables. The current pytables build (3.2.0) causes python to
        crash when you try to use the "blosc" compression. Until a fix comes out downgrade
        pytables to version 3.1.1
        This can be done with conda using the following code in your command line of choice

        >>> conda install pytables=3.1.1

        Then saying yes to the downgrade warnings.

    Parameters:
        data (pygeostat.DataFile or pandas.DataFrame): data to write out
        flname (str): Path (or name) of file to write out.

    Keyword Args:
        variables (List(str)): List of variables to write out if only a subset is desired.
        hdf_key (str): Name of the group to write the data to in hdf5 file. If appending then
            multiple "groups" can be written to the same file.
        mode (str): write mode. `r` read only, `w` write a new file (existing file will be
            overwritten), `a` append (if no file exists a new file will be created), `r+ similar
            to `a` but file must already exist
        hdf_append (bool): For Table formats, append the input data to the existing
        complevel (int): Compression level from 0 (least) to 9 (greatest). Default: 9
        complib (str): Compression Library. None, "zlib", or "blosc". See Note for warning.
            Default: "blosc"
        chuncksize (int): Write out by chunks using the chunksize. This can significantly
            lower your memory usage in writing.

    Things To Look Into Adding:

        1. In the future we might add the  keyword argument "expectedrows"
        which will allow pandas/pytables to optimize the read/write performance.

        2. Adding the attributes (like grid definitions!) to the HDF file

    To write out a new file

    >>> gs.write_hdf5(data, flname, hdf_key='group name')

    To append data to an already existing hdf5 file

    >>> gs.write_hdf5(data, flname, hdf_key='group name', mode='a')

    To append rows to an existing table in a hdf5 file

    >>> gs.write_hdf5(data, flname, hdf_key'group name in file to append to', mode='a',
    >>>               hdf_append=True)

    """
    from .data import DataFile as DataFile
    # If a DataFile is used, check the arguments
    if complevel is None:
        complevel = 9
    if complib is None:
        complib = "blosc"
    if isinstance(data, DataFile):
        if variables is None:
            variables = data.data.columns.tolist()
    else:
        if variables is None:
            variables = data.columns.tolist()
    # Force a write
    tempstore = pd.HDFStore(flname, mode=mode)
    if isinstance(data, DataFile):
        tempstore.append(hdf_key, data.data[variables], index=True, DataColumns=True,
                         complib=complib, complevel=complevel, append=hdf_append)
    else:
        tempstore.append(hdf_key, data[variables], index=True, DataColumns=True,
                         complib=complib, complevel=complevel, append=hdf_append)
    # if a griddef is available or passed then append the info to the file

    if griddef is not None:
        tempstore.root._v_attrs.gridstr = str(griddef)
    elif hasattr(data, 'griddef'):
        if hasattr(data, 'store'):
            try:
                data.store.root._v_attrs.gridstr
            except AttributeError:
                print('Warning: overwritting griddef in h5 file')
                tempstore.root._v_attrs.gridstr = str(data.griddef)
        else:
            tempstore.root._v_attrs.gridstr = str(data.griddef)

    tempstore.close()

def open_h5(flname):
    '''Open a hdf5 without reading any of the data into memory.

    Parameters:
        flname (str): Path (or name) of file to read.

    Returns:
        store (pandas.pytables.HDFStore): Pytables HDFStore object

    Quickly read in a HDF5 file and look at the available tables

    >>> store = gs.open_hdf5(flname)
    >>> store

    To create a DataFile object with no data and then review what tables are stored in the file

    >>> datafl = gs.DataFile(flname=filename, headeronly=True, hdf5_header=True)
    >>> datafl.store

    To load a table into the DataFile as well as the store info

    >>> datafl = gs.DataFile(flname=filename, hdf5_header=True)
    '''
    store = pd.HDFStore(flname)

    return store


def read_h5_pytables(flname, hdf_key='gslib_data'):
    '''Reads in a GSLIB-style HDF5 binary data file. This is useful for greatly increasing
    the read and write speads to memory which is a common bottleneck. There are
    some usage notes and warnings that should be reviewed.

    For more information on HDF5 - search youtube for some of the python hdf5 review videos.

    For a HDF5 file viewing software (outside of python) with a GUI interface check out
    the HDFView software at http://www.hdfgroup.org

    Parameters:
        flname (str): Path (or name) of file to read.

    Keyword Args:
        hdf_key (str): The name of the group in the file which the data is stored under

    Returns:
        data (pandas.DataFrame): Pandas DataFrame object with input data.

    Note:
        To open the file without reading any data into memory use ``gs.open_hdf5``
        This function uses the builtin functionality of pandas which in turn utilizes Pytables.
        Pytables is row oriented so it seems like you can't append more column to a current table
        once in the h5 file. You can append more rows though. You would need to overwrite the
        and create a new table if you want more columns.
        For some great discussion on large data workflows check out this stack overflow question.
        http://stackoverflow.com/questions/14262433/large-data-work-flows-using-pandas
    '''
    # Can the file be opened?
    try:
        with open(flname, 'r'):
            pass
    except FileNotFoundError:
        raise Exception(' '.join([__name__, 'File could not be opened:', flname]))
    # Read in the data
    data = pd.read_hdf(flname, key=hdf_key)
    # varnames = data.columns.tolist()

    # Only return the DataFrame
    return data


def convert_rlzns_to_h5(flname, writeflname, griddef=None, num_real=0, real_size=0, fltype=None, complevel=9, complib='blosc', label_padding=3, subfolder='BT_Units/', label_prefix='Realization_', quiet=False):
    """
    A function to convert a text based file of multiple realizations to the HDF5 format. Each
    realization will be saved under the model directory in the specified subfolder as a seperate
    table in the file.

    Parameters:
        flname (str): Name of the realization file to convert to hdf5
        writeflname (str): File name for output hdf5 file
        griddef (obj): pygeostat.GridDef grid definition
        num_real (int): Number of realizations in the file
        real_size (int): Length of each realization
        fltype (str): Type of file being read `gslib` or `csv`. default is `gslib`
        complevel (0-9): Level of compression in the hdf5 file. 0 = None, 9 = Most compressed
        comlib (str): compression library to use `blosc`, `zlib` etc. Can use any pytables
            compression library
        label_padding (int): Zero padding for creating the labels. By default it is set to three
            so the first realization will be labeled as `Realization_001`
        label_prefix (str): The group name and file structure for storing in the H5 file. By default
            this is set up as `Models/Realization_xxx`

    Note:
        Needs either a grid definiton or number of realizations.

        If number of realizations and realization size are not passed then it will perform a quick
        read of the file to determine how many rows of data are in the file and use this to figure
        out missing information

    Examples:
        A simple fast call:

        >>> gs.convert_rlzns_to_h5('sgsim.out', 'project.h5', griddef=grid, num_real=100)
        >>> '100 Realizations written to project.h5'
        >>> 'Realizations written to the following h5 directory: Models/BT_Units/'
    """
    import warnings
    from .datautils.labels.make_labels import make_labels as make_labels
    # ------------------------------------------------
    # Sanity checks
    # ------------------------------------------------
    # Check to make sure either a griddef is give or the number of realizations
    if griddef is not None:
        real_size = griddef.count()
        gridstr = str(griddef)
    if griddef is None:
        if num_real is 0:
            raise Exception('ERROR: Needs either a grid definition or the number of realizations')

    # If no filetype is given guess at what type of file
    if fltype is None:
        # Guess a file type based on the file name if obvious
        if flname.endswith('.out') or flname.endswith('.dat'):
            fltype = 'GSLIB'
        elif flname.endswith('.csv'):
            fltype = 'CSV'
        elif flname.endswith('.h5') or flname.endswith('.hdf5') or flname.endswith('.hd5'):
            fltype = 'HDF5'
        else:
            # Otherwise just assume GSLIB for now but this may change
            fltype = 'GSLIB'

    # Raise an exception if the file is already a hdf5 file
    if fltype.lower() == 'hdf5':
        raise Exception('Not currently set up to handle hdf5 as INPUT file. Only as OUTPUT file')

    # ------------------------------------------------
    # Process header information from the files
    # ------------------------------------------------
    # If gslib style then skip the header info lines
    if fltype.lower() == 'gslib':
        with open(flname, 'r') as datafl:
            # Geo-EAS Header
            _ = datafl.readline().strip()  # Title
            nvar = int(datafl.readline().split()[0])
            varnames = []
            for _ in range(nvar):
                varnames.append(datafl.readline().strip())
            nrowtoskip = 2 + nvar
            header_lines = nrowtoskip
        delimiter = r'\s*'
        header = None

    # Set read options if csv type file
    if fltype.lower() == 'csv':
        temp = pd.read_csv(flname, header=0, nrows=1)
        varnames = temp.columns.tolist()
        del temp
        nrowtoskip = 0
        header_lines = 1
        delimiter = ','
        header = 0

    # ------------------------------------------------
    # Set initial options and find any missing info
    # ------------------------------------------------

    # If either real_size or num_real is missing figure out what they are
    if real_size is 0:
        lines = iotools.file_nlines(flname)
        data_size = lines - header_lines
        real_size = int(data_size / num_real)
    if num_real is 0:
        lines = iotools.file_nlines(flname)
        data_size = lines - header_lines
        num_real = int(data_size / real_size)
    # a couple of random variables including a list of group names
    len_var = len(varnames)
    # This allows us to be dumb and will still work if we forget how many realizations we have
    if num_real < 1000:
        num_labels = 1000
    else:
        num_labels = num_real
        if label_padding < 4:  # increase the padding if we have more than 1000 realizations
            label_padding = 4
    table_name = make_labels(label_prefix, num_labels, padding=label_padding)
    idx = 0
    # ------------------------------------------------
    # Iterratively write to the hdf5 file
    # ------------------------------------------------

    # Open a HDFfile to write to. Note will overwrite any existing file
    tempstore = pd.HDFStore(writeflname, 'w')

    # Because we can.... Write some grid information to the hdf5 file
    tempstore.root._v_attrs.realsize = real_size
    tempstore.root._v_attrs.numreal = num_real
    if gridstr:
        tempstore.root._v_attrs.gridstr = gridstr

    # Iterate through the read file chunk by chunk and write it to a hdf5 file
    for chunk in pd.read_csv(flname, skiprows=nrowtoskip, skipinitialspace=True,
                             chunksize=real_size, header=header, engine='python',
                             delimiter=delimiter):
        # Check to see if any trailing spaces got read in as a new column
        len_ch_col = len(chunk.columns)
        if len_ch_col > len_var:
            didx = len(varnames)
            del chunk[didx]
            warnings.warn('There is probably trailing spaces in the GSLIB file. Found ', len_ch_col,
                          'columns, but there is header information for only ', len_var,
                          'columns. Last column was assumed to be trailing spaces and '
                          'therefore was deleted')
        # Set column names and write to h5 file
        chunk.columns = varnames
        tempstore.append(table_name[idx], chunk, header=False, index=True, DataColumns=True,
                         complevel=complevel, complib=complib)
        idx += 1

    # Check to make sure we wrote the correct number of realizations to file
    tables_appended = idx - 1
    if tables_appended != num_real:
        print('WARNING: Number of realizations =', num_real,
              ' But number of realizations appended to h5 file =', tables_appended)

    # ------------------------------------------------
    # Clean up after ourselves
    # ------------------------------------------------
    tempstore.close()

    if quiet is False:
        print((idx - 1), ' Realizations written to ', flname)
        print('Realizations written to the following h5 directory: Models/' + subfolder)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Pit plotting routine with 2D slices using matplotlib"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------


from matplotlib.path import Path
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np

from . set_style import set_plot_style
from .. pygeostat_parameters import Parameters


@set_plot_style
def pit_plot(arr, griddef, ax=None, orient='xz', slice_number=0, lineweight=1, color='k', iso=0.5,
            linestyle='solid', figsize=None, xlim=None, ylim=None, title=None, xlabel=None,
            ylabel=None, unit=None, rotateticks=None, randomize=False, grid=None, axis_xy=None,
            cust_style=None, label=None, output_file=None, out_kws=None):
    """This funcion will take an array of indicators (from lg3d) and an orientation, and plot the
    pit shell lines for a given cross section view.

    Note:
        This function can only deal with 1 realization in the file
        so if you have multiple realizations you need to either pass a slice to this function or copy 1
        realization to a separate file.

    Parameters:
        arr (Array): Array (DataFrame Column) passed to the program with indicator
            values (i.e. 0 & 1)
        ax (mpl.axis): Matplotlib axis to plot the figure
        orient (str): Orientation to slice data. 'xy', 'xz', 'yz' are the only accepted values
        slice_number (int): Location of slice to plot
        lineweight (float): Any Matplotlib line weight
        color (str): Any Matplotlib color
        iso (float): Inside or Outside of Pit limit (i.e. if greater than 0.5 inside of pit)
        linestyle (str): Any Matplotlib linestyle
        randomize (bool): True or False... obviously
        figsize (tuple): Figure size (width, height)
        title (str): Title for the plot
        xlabel (str): X-axis label
        ylabel (str): Y-axis label
        unit (str): Unit to place inside the axis label parentheses
        rotateticks (bool tuple): Indicate if the axis tick labels should be rotated (x, y)
        grid (bool): Plots the major grid lines if True. Based on Parameters['plotting.grid']
            if None.
        axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
            visible) if axis_xy is True. Based on Parameters['plotting.axis_xy'] if None.
        label (str): Legend label to be added to Matplotlib axis artist
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`


    Returns:
        fig (fig): Matplotlib figure instance

    Examples:
        A simple call:

        >>> gs.pit_plot(data.data, data.griddef, title='Pit Outline Using LG3D output')

        .. image:: ./figures/PlottingGallery/pitplot.png

        In order to plot multiple pits (say from a file with multiple realizations) you have can
        plot to the same matplotlib axis. For multiple realizations using a loop is the easiest
        as shown below.

        >>> sim = gs.DataFile(flname='SGS.lg3d', griddef=grid_5m)

        Loop through the SGSIM LG3D output file

        First plot the first realization and grab the matplotlib axis

        >>> import matplotlib.pyplt as plt
        ... rmin = 0
        ... rmax = pit.griddef.count()
        ... fig = gs.pit_plot(sim.data[rmin:rmax], sim.griddef, title='Pit Outline Using LG3D output
        ...                 with Multiple Realizations')
        ... ax = fig.gca()

        Then loop through the rest of the realizations (Say 50) and plot them on current axis

        >>> for i in range (1, 50):
        ...     rmin = i*pit.griddef.count()
        ...     rmax = rmin + pit.griddef.count()
        ...     gs.pit_plot(sim.data[rmin:rmax], sim.griddef, ax=ax)

        Save the figure

        >>> gs.export_image('pitplot_mr.png', format='png')

        .. image:: ./figures/PlottingGallery/pitplot_mr.png
    """
    from .export_image import export_image
    from . utils import format_plot, _spatial_labels, _tickoverlap
    
    # Handle dictionary defaults
    if out_kws is None:
        out_kws = dict()
    # Set-up plot if no axis is supplied
    if ax is None:
        fig, ax = plt.subplots(1, figsize=figsize)
    # Handle panda input
    if hasattr(arr, 'values'):
        arr = arr.values

    if arr.shape != (griddef.nz, griddef.ny, griddef.nx):
        arr = arr.reshape((griddef.nz, griddef.ny, griddef.nx))

    binarr = np.zeros_like(arr, dtype=bool)
    binarr[arr > iso] = True

    if orient == 'xy':
        # view = binarr[slice_number, :, :]
        print('Orientation of ', orient, ' not yet supported - to be added!')
        return
    elif orient == 'xz':
        view = binarr[:, slice_number, :]
        # Get limits for orient
    elif orient == 'yz':
        view = binarr[:, :, slice_number]
        # Get limits for orient
    else:
        print('Orientation of ', orient, ' not supported')
        return

    a = orient[0]
    b = orient[1]

    xmin = getattr(griddef, a + 'limits')[0]
    xmax = getattr(griddef, a + 'limits')[1]
    ymin = getattr(griddef, b + 'limits')[0]
    ymax = getattr(griddef, b + 'limits')[1]

    x = []
    y = []

    xmn = getattr(griddef, a + 'mn')
    nx = getattr(griddef, 'n' + a)
    xsiz = getattr(griddef, a + 'siz')

    for i in range(1, nx - 1):
        for v, k in zip(view[:, i], range(griddef.nz)):
            if v:
                x = x + [xmn + i * xsiz]
                y = y + [griddef.zmn + k * griddef.zsiz - griddef.zsiz / 2]
                break
            if k == griddef.nz - 1:
                if view[k, i + 1] or view[k, i - 1]:
                    x = x + [xmn + i * xsiz]
                    y = y + [griddef.zmn + k * griddef.zsiz + griddef.zsiz / 2]

    if randomize:
        r = (np.random.random() - 0.5) * xsiz
        for i in range(len(x)):
            x[i] = x[i] + r

    if len(x) > 2:
        verts = list(zip(x, y))
        codes = [Path.MOVETO] + [Path.LINETO] * (len(x) - 1)

        path = Path(verts, codes)
        patch = patches.PathPatch(path, facecolor='none', ec=color, lw=lineweight, ls=linestyle)

        ax.add_patch(patch)
        fig = plt.gcf()
    # Add Legend artist to axis if label is passed
    if label:
        line1, = ax.plot([-9999999, -9999999, -9999999], label=label, linestyle=linestyle,
                         linewidth=lineweight, color=color)
        line1.set_label(label)
    # Set axis limits
    if xlim is None:
        xlim = (xmin, xmax)
    if ylim is None:
        ylim = (ymin, ymax)
    ax = _spatial_labels(ax, orient, griddef, slice_number, title, xlabel, ylabel, unit)
    if axis_xy is None:
        axis_xy = Parameters['plotting.axis_xy_spatial']
    format_plot(ax, xlim=xlim, ylim=ylim, grid=grid, axis_xy=axis_xy)
    # Note on Tick Labels:
    #   If a group of subplots are put together which share a x-axis, the rotation may not work. By
    #   getting the tick labels generated by matplotlib as a set of label objects, they can be
    #   looped through and have their settings individually fixed. This appears to be the only way
    #   to have the shared axis labels formated properly. The labels are also adjusted closer to the
    #   axis for esthetics. --Warren E. Black

    # The plots tick labels will not be properly accessible until the figure is "drawn", once the
    # command below is run, ax.get_ticklabel() will actually work properly.
    plt.draw()
    # Check to see if the ytick labels need to be rotated if the rotate argument was not passed.
    # This doesn't work if the figure is being placed into a subplot (i.e., is not a standalone
    # figure). Don't know why...
    if rotateticks is None:
        rotateticks = Parameters['plotting.rotateticks']
    if rotateticks is None:
        rotateticks = _tickoverlap(ax)
    # Configure y-axis tick labels
    ylabels = ax.get_yticklabels()
    for ylabel in ylabels:
        ylabel.set_ha('right')
        ylabel.set_va('center')
        # Rotate if required
        if rotateticks[1]:
            ylabel.set_rotation(90)
    # Configure x-axis tick labels
    xlabels = ax.get_xticklabels()
    for xlabel in xlabels:
        xlabel.set_ha('center')
        xlabel.set_va('top')
        # Rotate if required
        if rotateticks[0]:
            xlabel.set_rotation(45)
    # Fix tick label padding
    ax.tick_params(axis='both', pad=2)
    if rotateticks[1]:
        ax.tick_params(axis='y', pad=0)
    # Export figure
    if output_file or ('pdfpages' in out_kws):
        export_image(output_file, **out_kws)
    if ax is None:
        return fig
    else:
        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A basic probability plotting routine reminiscent of probability_plot from gslib"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from . set_style import set_plot_style


@set_plot_style
def probability_plot(data, wt=None, lower=None, upper=None, logscale=True, ax=None, figsize=None,
                    xlim=None, ylim=(0.01, 99.99), nyticks=15, title=None, xlabel=None, s=2, color='k',
                    grid=True, axis_xy=None, line=False, plot_style=None, custom_style=None, output_file=None,
                    out_kws=None, **kwargs):
    """
    Create either a normal or a lognormal probability plot. This plot displays all the data
    values on a chart that illustrates the general distribution shape and the behavior of the
    extreme values. Please review the documentation of the :func:`gs.set_style()
    <pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
    <pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
    their use in this function can be understood.

    This function requires the python package probscale. It can be installed be executing the
    following code in the command prompt:

        >>> pip install probscale

    Parameters:
        data: Tidy (long-form) 1D data where a single column of the variable exists with each row
            is an observation. A pandas dataframe/series or numpy array can be passed.
        wt: 1D dataframe, series, or numpy array of declustering weights for the data.
        lower (float): Lower trimming limits
        upper (float): Upper trimming limits
        ax (mpl.axis): Matplotlib axis to plot the figure
        figsize (tuple): Figure size (width, height)
        xlim (float tuple): Minimum and maximum limits of data along the x axis
        ylim (float tuple): Minimum and maximum limits of data along the y axis, e.g.(0.001, 99.999)
        nyticks (int): the number of ticks on the y axis to show. Currently disabled due to altered
            matplotlib functionality.
        title (str): Title for the plot
        xlabel (str): X-axis label. A default value of ``None`` will try and grab a label from the
            passed ``data``. Pass ``False`` to not have an xlabel.
        s (int): Size of points
        color (str or int): Any permissible matplotlib color or a integer which is used to draw a
            color from the pygeostat color pallet ``pallet_pastel`` (useful for iteration)
        grid(bool): plots the major grid lines if True. Based on gsParams['plotting.grid']
            if None.
        axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
            visible) if axis_xy is True. Based on gsParams['plotting.axis_xy'] if None.
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
        **kwargs: Optional permissible keyword arguments to pass to matplotlib's scatter function

    Returns:
        ax (ax): matplotlib Axes object with the histogram

    **Examples:**

        A simple call:

        .. plot::

            import pygeostat as gs
            data = gs.ExampleData('oilsands')
            gs.probability_plot(data.data['Fines'], logscale=False)

    """
    # Import the rest of the packages
    import numpy as np
    import pandas as pd
    from .export_image import export_image
    from .. statistics.cdf import cdf
    import matplotlib.pyplot as plt
    from .utils import format_plot, get_label
    try:
        import probscale
    except ImportError:
        raise ImportError('Looks like probscale is not installed, use:\n' +
                          '>>> pip install probscale\n')
    # Ensure valid VTK option
    # Coerce the passed data and wt into a single pandas dataframe
    if xlabel is None:
        xlabel = get_label(data)
    try:
        data = pd.DataFrame(data=data)
        data.reset_index(inplace=True, drop=True)
        if data.shape[1] != 1:
            raise ValueError("The passed `data` is not 1-D")
    except:
        raise ValueError("Please ensure the passed `data` can be coerced into a pandas dataframe"
                         " (i.e., data = pd.DataFrame(data=data)")
    if wt is not None:
        try:
            wt = pd.DataFrame(data=wt)
            if wt.shape[1] != 1:
                raise ValueError("The passed `wt` is not 1-D")
        except:
            raise ValueError("Please ensure the passed `wt` can be coerced into a pandas"
                             " dataframe (i.e., wt = pd.DataFrame(wt=wt)")
    else:
        wt = pd.DataFrame(data=np.ones((len(data), 1)))
    data = pd.concat([data, wt], axis=1)
    data.columns = ['data', 'wt']
    # Trim the data
    if lower is None:
        lower = data['data'].min()
    if upper is None:
        upper = data['data'].max()
    data.query("%s <= data <= %s" % (lower, upper))
    # Sanity checks
    assert((data['wt']).all() >= 0.0), 'Weights less than 0 not valid'
    # Handle dictionary defaults
    if out_kws is None:
        out_kws = dict()
    # Calculate the cdf
    cdf_x, cdfvals = cdf(data['data'], weights=data['wt'])
    cdfvals = cdfvals * 100
    # Set-up plot if no axis is supplied
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    # Plot the figure
    if line:
        ax.plot(cdf_x, cdfvals, color=color, **kwargs)
    else:
        ax.scatter(cdf_x, cdfvals, s=s, c=color, lw=0, **kwargs)
    # Set some figure aesthetics
    if logscale:
        ax.set_xscale('log')
    ax.set_yscale('prob')
    ax.locator_params(axis='y')
    ax = format_plot(ax, xlabel=xlabel, ylabel='Cumulative Probability', title=title,
                   grid=grid, axis_xy=axis_xy, xlim=xlim, ylim=ylim)
    # Export figure
    if output_file or ('pdfpages' in out_kws):
        export_image(output_file, **out_kws)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""validation_plot plots a scatter plot of cross validation results with relevant statistics"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import numpy as np
import pandas as pd
from scipy import stats
import matplotlib as mpl
import matplotlib.pyplot as plt
from .utils import get_label, get_statblk
from .export_image import export_image
from ..statistics.kde import kde_scipy
from . set_style import set_plot_style


@set_plot_style
def validation_plot(x, y, figsize=None, vlim=None, xlabel=None, ylabel=None, title=None, stat_blk='all',
                    stat_xy=(0.95, 0.05), stat_ha=None, stat_fontsize=None, mc='k', ms=None, plot_style=None,
                    lw=None, grid=None, axis_xy=None, custom_style=None, output_file=None, ax=None, dens=False, 
                    rasterized=False, **kwargs):
    """
    This function uses numpy to calculate the regression model and matplotlib to plot the
    scatter plot, regression line, and 45 degree line. Statistics are calculated using numpy.

    The only parameters needed are the ``x`` and ``y``. All of the other arguments are optional. If
    the label parameters are left to their default value of ``None``, the column information will
    be used to label the axes.

    Two statistics block sets are available: ``'minimal'`` and the default ``'all'``. The
    statistics block can be customized to a user defined list and order. Available statistics are
    as follows:

    >>> ['ndat', 'ymean', 'ystdev', 'xmean', 'xstdev', 'cov', 'rho', 'mse', 'sor']

    Please review the documentation of the :func:`gs.set_style()
    <pygeostat.plotting.set_style.set_style>` and :func:`gs.export_image()
    <pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
    their use in this function can be understood.

    Parameters:
        x: Tidy (long-form) 1D data where a single column of the variable to plot along the x-axis
            exists with each row is an observation. A pandas dataframe/series or numpy array can be
            passed.
        y: Tidy (long-form) 1D data where a single column of the variable to plot along the y-axis
            exists with each row is an observation. A pandas dataframe/series or numpy array can be
            passed.
        figsize (tuple): Figure size (width, height)
        vlim (float tuple): A single tuple for the minimum and maximum limits of data along both
            axes. Will not be a symmetrical plot if they are not the same value
        xlabel (str): X-axis label
        ylabel (str): Y-axis label
        title (str): Title for the plot
        stat_blk (str or list): Indicate what preset statistics block to write or a specific list
        stat_xy (str or float tuple): X, Y coordinates of the annotated statistics in figure
            space.
        stat_ha (str): Horizontal alignment parameter for the annotated statistics. Can be
            ``'right'``, ``'left'``, or ``'center'``. The value ``None`` can also be used to allow
            the parameter ``stat_xy`` to determine the alignment automatically.
        stat_fontsize (float): the fontsize for the statistics block. If None, based on
            gsParams['plotting.stat_fontsize']. If less than 1, it is the fraction of the
            matplotlib.rcParams['font.size']. If greater than 1, it the absolute font size.
        mc (str): Any permissible matplotlib color value for the scatter plot markers
        ms (float): Size of scatter plot markers
        grid(bool): plots the major grid lines if True. Based on gsParams['plotting.grid']
            if None.
        axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
            visible) if axis_xy is True. Based on gsParams['plotting.axis_xy'] if None.
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        **kwargs: Optional permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`

    Returns:
        ax (ax): Matplotlib Axes object with the cross validation plot

    **Examples:**
    
        A simple call:

        .. plot::

            import pygeostat as gs
            data = gs.ExampleData('3d_estimate')
            gs.validation_plot(data.data['Estimate'], data.data['True'], stat_blk='minimal')

        |

        Fixing the value limits, moving the statistics block, and exporting the figure.

        .. plot::

            import pygeostat as gs
            import numpy as np
            mean = [0, 0]
            cov = [[1, 0.8], [0.8, 1]]  # diagonal covariance
            x, y = np.random.multivariate_normal(mean, cov, 5000).T
            gs.validation_plot(x,y,vlim=(-3.5, 3.5) ,grid=True, stat_xy=(1, 0.68))

    |

    .. codeauthor:: pygeostat development team 2015-08-05
    """
    import types
    from .utils import format_plot, _set_stat_fontsize
    # Convert list to pd.series
    if isinstance(x, list):
        x = pd.DataFrame(x)
        x = x[0]
    if isinstance(y, list):
        y = pd.DataFrame(y)
        y = y[0]
    if isinstance(x, pd.DataFrame):
        x = x[x.columns[0]]
    if isinstance(y, pd.DataFrame):
        y = y[y.columns[0]]
    # Set-up figure
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    # Configure plot
    ax.axis('equal')
    if xlabel is None:
        xlabel = get_label(x)
        if xlabel == 0:
            xlabel = None
        if xlabel is None:
            xlabel = 'Estimate'
    if ylabel is None:
        ylabel = get_label(y)
        if ylabel == 0:
            ylabel = None
        if ylabel is None:
            ylabel = 'True'
    ax = format_plot(ax, xlabel=xlabel, ylabel=ylabel, title=title, grid=grid, axis_xy=axis_xy, xlim=vlim, ylim=vlim)
    #ax.tick_params(axis='both', pad=2)
    if isinstance(dens, types.FunctionType):  # fucntion takes (x, y, #comps) as positional args and returns `colors`
        mc = dens(x, y, 8)
    else:
        if dens:
            # plot the points as density colors
            values = np.vstack([x, y])
            mc = kde_scipy(values, values)
    if lw is None:
        lw = 0
    # Plot the scatter plot
    ax.scatter(x, y, c=mc, s=ms, lw=lw, rasterized=rasterized)
    # Get the axis limits determined by plt.scatter()
    if ax.get_ylim()[1] >= ax.get_xlim()[1]:
        maxlim = ax.get_ylim()
    else:
        maxlim = ax.get_xlim()
    # Plot the regression model
    regmodel = np.polyfit(x, y, 1)
    predicted = np.polyval(regmodel, [maxlim[0], maxlim[1]])
    ax.plot([maxlim[0], maxlim[1]], predicted, '-', lw=0.5, color='red')
    # Plots stats
    if stat_blk:
        ndat = len(x)
        ymean = np.nanmean(y)
        ystdev = np.nanstd(y)
        xmean = np.nanmean(x)
        xstdev = np.nanstd(x)
        cov = np.cov(x, y)[0, 1]
        rho = stats.pearsonr(x, y)[0]
        mse = np.sqrt(((x - y) ** 2).mean())
        sor = stats.linregress(x, y)[0]
        statlist = {'ndat': ('$n = {}$'.format(ndat)),
                    'ymean': (r'$m_y = {:0.3f}$'.format(ymean)),
                    'ystdev': (r'$\sigma_y = {:0.3f}$'.format(ystdev)),
                    'xmean': (r'$m_x = {:0.3f}$'.format(xmean)),
                    'xstdev': (r'$\sigma_x = {:0.3f}$'.format(xstdev)),
                    'cov': ('$cov = {:0.3f}$'.format(cov)),
                    'rho': (r'$\rho = {:0.3f}$'.format(rho)),
                    'mse': ('$RMSE = {:0.3f}$'.format(mse)),
                    'sor': ('$SoR = {:0.3f}$'.format(sor)),
                    'newline': ''}
        # Default statistic sets
        statsets = {'minimal': ['ndat', 'rho', 'mse', 'sor'],
                    'all': ['ndat', 'newline', 'ymean', 'ystdev', 'newline', 'xmean', 'xstdev',
                            'newline', 'cov', 'rho', 'mse', 'sor'],
                    'none': None}
        txtstats, stat_xy, ha, va = get_statblk(stat_blk, statsets, statlist, stat_xy)
        if stat_ha:
            ha = stat_ha
        stat_fontsize = _set_stat_fontsize(stat_fontsize)
        ax.text(stat_xy[0], stat_xy[1], txtstats, va=va, ha=ha, fontsize=stat_fontsize,
                transform=ax.transAxes)
    # Plot the 45 degree line
    ax.plot(maxlim, maxlim, lw=0.5, color='k')
    ax.set(xlim=maxlim, ylim=maxlim)  # Reset the axis limits as the 45 degree line pushed it out

    # Export figure
    if output_file or ('pdfpages' in kwargs):
        export_image(output_file, **kwargs)

    return ax
